# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Dict

from qctrlcommons.exceptions import QctrlArgumentsValueError

from fireopal._event_tracker import (
    _EVENT_TRACKER,
    UserTraits,
    _check_for_remote_router,
    _get_user_payload,
)

Credentials = Dict[str, str]


def make_credentials_for_ibmq(
    token: str, group: str, hub: str, project: str
) -> Credentials:
    """
    Make a Credentials dictionary for IBM Quantum.

    Parameters
    ----------
    token : str
        The IBM Quantum account token.
    group : str
        The IBM Quantum group.
    hub : str
        The IBM Quantum hub.
    project : str
        The IBM Quantum project.

    Returns
    -------
    Credentials
        A dictionary usable for the `credentials`
        argument of any Fire Opal function using IBM Quantum.

    Notes
    -----
    This function performs only basic type checking of
    the credentials it receives. It does not check whether
    the credentials are valid for hardware access.
    """
    if _check_for_remote_router():
        user_traits = UserTraits(_get_user_payload())
        _EVENT_TRACKER.track_make_credentials_for_ibmq(user_traits=user_traits)
    _check_all_strings(token=token, group=group, hub=hub, project=project)
    return {
        "token": token,
        "group": group,
        "hub": hub,
        "project": project,
        "provider": "ibmq",
    }


def make_credentials_for_ibm_cloud(token: str, instance: str) -> Credentials:
    """
    The Credentials builder for IBM cloud devices.

    Parameters
    ----------
    token : str
        The IBM API account token.
    instance : str,
        The IBM cloud instance, the user's CRN.

    Returns
    -------
    Credentials
        A dictionary usable for the `credentials`
        argument of any Fire Opal function using IBM Cloud.
    """
    if _check_for_remote_router():
        user_traits = UserTraits(_get_user_payload())
        _EVENT_TRACKER.track_make_credentials_for_ibm_cloud(user_traits=user_traits)
    _check_all_strings(token=token, instance=instance)
    return {"token": token, "instance": instance, "provider": "ibm_cloud"}


def make_credentials_for_braket(arn: str) -> Credentials:
    """
    Make a Credentials dictionary for Braket.

    Parameters
    ----------
    arn : str
        The Amazon resource number for an IAM role
        that has Braket permissions and trusts Q-CTRL's
        AWS account.

    Returns
    -------
    Credentials
        A dictionary usable for the `credentials`
        argument of any Fire Opal web API function.

    Notes
    -----
    This function performs only basic type checking of
    the credentials it receives. It does not check whether
    the credentials are valid for hardware access.
    """
    _check_all_strings(arn=arn)
    return {"arn": arn, "provider": "braket"}


def _check_all_strings(**kwargs: str) -> None:
    """
    Check whether all arguments are strings.

    Raises
    ------
    QctrlArgumentsValueError
        If there are non-string arguments.
    """
    if not all((isinstance(arg, str) for arg in kwargs.values())):
        raise QctrlArgumentsValueError(
            description="All arguments should be strings.",
            arguments={f"{key} type": type(value) for key, value in kwargs.items()},
        )



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import Dict

from .base import fire_opal_workflow


@fire_opal_workflow("read_database_workflow")
def read_data(ibm_device_name: str, entry: str) -> Dict:
    """
    Reads an entry from the Fire Opal database.

    Parameters
    ----------
    ibm_device_name : str
        The name of the ibm device associated with the desired data.
    entry : str
        The database entry label.

    Returns
    -------
    Dict
        The output of the read database workflow.
    """
    return {"ibm_device_name": ibm_device_name, "entry": entry}



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

__version__ = "8.1.0"

from qctrlworkflowclient.utils import check_package_version as _check_package_version

from . import credentials
from ._utils import print_package_versions
from .config import configure_organization
from .constants import PACKAGE_INFO as _package_info
from .functions import (
    activity_monitor,
    authenticate_qctrl_account,
    execute,
    get_action_metadata,
    get_result,
    iterate,
    show_supported_devices,
    solve_qaoa,
    stop_iterate,
    validate,
)

_check_package_version(_package_info)

# Note: Need to include the late import below for
# docs to be generated, while avoiding a circular
# import issue.
# pylint: disable=wrong-import-position
from .fire_opal_job import FireOpalJob

__all__ = [
    "activity_monitor",
    "authenticate_qctrl_account",
    "credentials",
    "configure_organization",
    "execute",
    "get_action_metadata",
    "get_result",
    "iterate",
    "show_supported_devices",
    "solve_qaoa",
    "stop_iterate",
    "validate",
    "print_package_versions",
    "FireOpalJob",
]



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Fire Opal Job class."""

from __future__ import annotations

from typing import Any

from qctrlworkflowclient.router.api import (
    Action,
    ActionStatus,
    ApiRouter,
)

from fireopal._event_tracker import (
    _EVENT_TRACKER,
    UserTraits,
    _check_for_remote_router,
    _get_user_payload,
)
from fireopal.config import get_config
from fireopal.functions.get_result import get_result

_FIRE_OPAL_JOB_STATUSES = {
    ActionStatus.SUCCESS.value: "Job has finished successfully.",
    ActionStatus.REVOKED.value: "Job has been cancelled.",
    ActionStatus.FAILURE.value: "Job has failed.",
    ActionStatus.PENDING.value: "Job has been submitted to Q-CTRL.",
    ActionStatus.RECEIVED.value: "Job has been submitted to Q-CTRL.",
    ActionStatus.RETRY.value: "Job has been submitted to Q-CTRL.",
    ActionStatus.STARTED.value: "Job has been submitted to Q-CTRL.",
}


class FireOpalJob:
    """
    Fire Opal Job class.
    """

    def __init__(self, action: Action) -> None:
        """
        Contains information regarding jobs submitted to Q-CTRL.

        Parameters
        ----------
        action : Action
            The unique identifier object for a job submitted
            to Q-CTRL.
        """
        self._action = action

    @property
    def action_id(self) -> str:
        """
        The identifier for a Q-CTRL job.

        Returns
        -------
        str
            The identifier for a Q-CTRL job.
        """
        return self._action.action_id

    def status(self) -> dict[str, str]:
        """
        The status of the Q-CTRL Fire Opal job.
        """
        router: ApiRouter = get_config().get_router()
        if _check_for_remote_router():
            user_traits = UserTraits(_get_user_payload(router))
            _EVENT_TRACKER.track_fire_opal_job_status(
                user_traits=user_traits, action_id=self.action_id
            )

        assert isinstance(router, ApiRouter)
        action = router.update_action_status(action=self._action)

        return {
            "status_message": _FIRE_OPAL_JOB_STATUSES[action.status],
            "action_status": action.status,
        }

    def result(self) -> dict[str, Any]:
        """
        The results of the Q-CTRL Fire Opal job.

        Returns
        -------
        dict[str, Any]
            The results of the Q-CTRL Fire Opal job.
        """
        if _check_for_remote_router():
            user_traits = UserTraits(_get_user_payload())
            _EVENT_TRACKER.track_fire_opal_job_result(
                user_traits=user_traits, action_id=self.action_id
            )
        return get_result(self._action.action_id)



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from qctrlworkflowclient.utils import PackageInfo as _PackageInfo

INVALID_SUBSCRIPTION_ERROR = """
---------------------------------------------------------------------
Your authentication succeeded, but your subscription is invalid for
this product.

Please access `https://fire.q-ctrl.com` for more details.
---------------------------------------------------------------------
"""

FIRE_OPAL_USER_ROLE = "fire-opal-cli-access"

API_KEY_NAME = "QCTRL_API_KEY"

PACKAGE_INFO = _PackageInfo(
    name="Fire Opal client",
    install_name="fire-opal",
    import_name="fireopal",
    changelog_url="https://docs.q-ctrl.com/fire-opal/changelog",
)



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from functools import partial
from typing import (
    Any,
    Callable,
)

from qctrlworkflowclient import (
    core_workflow,
    print_warnings,
)
from qctrlworkflowclient.functions import async_core_workflow
from qctrlworkflowclient.router.api import (
    Action,
    DecodedResult,
)

from fireopal.config import get_config
from fireopal.fire_opal_job import FireOpalJob


def _formatter(input_: DecodedResult) -> Any:
    result = input_.decoded
    if isinstance(result, dict):
        return print_warnings(result)
    return result


def _async_formatter(input_: dict[str, Action]) -> FireOpalJob:
    assert isinstance(input_, dict) and input_.get("async_result") is not None
    return FireOpalJob(action=input_["async_result"])


fire_opal_workflow: Callable = partial(core_workflow, get_config, formatter=_formatter)
async_fire_opal_workflow: Callable = partial(
    async_core_workflow, get_config, formatter=_async_formatter
)



# Copyright 2024 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import Optional

from fireopal._event_tracker import (
    _EVENT_TRACKER,
    UserTraits,
    _check_for_remote_router,
    _get_user_payload,
)
from fireopal._utils import log_activity
from fireopal.config import get_config
from fireopal.credentials import Credentials
from fireopal.functions._utils import (
    check_circuit_submission_validity,
    handle_single_item,
)

from ..fire_opal_job import FireOpalJob
from .base import async_fire_opal_workflow


@async_fire_opal_workflow("iterate_workflow")
def iterate(
    circuits: str | list[str],
    shot_count: int,
    credentials: Credentials,
    backend_name: str,
    parameters: Optional[dict[str, float] | list[dict[str, float]]] = None,
) -> FireOpalJob:
    """
    Submit a batch of `circuits` where `shot_count` measurements are taken per circuit.
    If permitted by the hardware provider, multiple calls to this function will
    re-use the same circuit submission queue position.

    Parameters
    ----------
    circuits : str or list[str]
        Quantum circuit(s) in the form of QASM strings. You may use Qiskit to
        generate these strings.
    shot_count : int
        Number of bitstrings that are sampled from the final quantum state.
    credentials : Credentials
        The credentials for running circuits. See the `credentials` module for functions
        to generate credentials for your desired provider.
    backend_name : str
        The backend device name that should be used to run circuits.
    parameters : dict[str, float] or list[dict[str, float]] or None, optional
        The list of parameters for the circuit(s), if they're parametric.
        Defaults to None.

    Returns
    -------
    FireOpalJob
        A job object containing the results of the execution.
    """
    log_activity(
        function_called="iterate",
        circuits=circuits,
        shot_count=shot_count,
        backend_name=backend_name,
        parameters=parameters,
    )

    if _check_for_remote_router():
        user_traits = UserTraits(_get_user_payload())
        _EVENT_TRACKER.track_iterate(
            user_traits=user_traits,
            circuit_count=len(circuits),
            parameter_count=len(parameters) if parameters else 0,
        )

    check_circuit_submission_validity(
        circuits=circuits,
        shot_count=shot_count,
        credentials=credentials,
        backend_name=backend_name,
        parameters=parameters,
    )
    circuits = handle_single_item(circuits)
    if parameters:
        parameters = handle_single_item(parameters)

    settings = get_config()
    credentials_with_org = credentials.copy()
    credentials_with_org.update({"organization": settings.organization})
    return {
        "circuits": circuits,
        "shot_count": shot_count,
        "credentials": credentials_with_org,
        "backend_name": backend_name,
        "parameters": parameters,
    }  # type: ignore[return-value]
